from collections import deque, abc
from concurrent import futures

from scrapgo.core import SoupParser, RequestsBase, CachedRequests
from scrapgo.core.http import RequestsBase
from scrapgo.lib import pluralize

from .actions import resolve_link, ReduceMixin



class CrawlMixin(ReduceMixin, SoupParser):
    urlorders = None

    def crawl(self, _urlorders=None, _results=None, parent_response=None, _composed:dict=None, _context:dict=None, _composed_buffer:dict=None):

        if _urlorders is None:
            action, *rest = self.urlorders
        elif len(_urlorders) == 0:
            return 
        else:
            action, *rest = _urlorders
        
        _context = _context or dict()
        
        _composed = _composed or dict()
        _composed.setdefault(action.name, [])

        _composed_buffer = _composed_buffer or dict()
        _composed_buffer.setdefault(action.name, [])
        

        def build_requests():
            def __resolve(link):
                url = resolve_link(action, link, parent_response)
                return self.dispatch_fields(action, url)

            def _build_urls():
                links = self.dispatch_renderer(action, _results, parent_response, _context)
                for link in links:
                    url = __resolve(link)
                    if self.dispatch_urlfilter(action, url, _results, _context):
                        yield url

            def _build_payloads():
                yield from self.dispatch_payloader(action, _results, _context)
                
            def __build_kwargs(url, payload):
                cookies = self.dispatch_cookies(action)
                headers = self.dispatch_headers(action)
                payload = payload or None
                kwargs = dict(url=url, headers=headers, cookies=cookies)
                json_type = False
                if content_type := headers.get('Content-Type'):
                    if 'application/json' in content_type:
                        json_type = isinstance(payload, (dict, list))
                kwargs['json' if json_type else 'data'] = payload
                return kwargs

            for url in _build_urls():
                for payload in _build_payloads():
                    kwargs = __build_kwargs(url, payload)
                    yield kwargs
        
        def process_response(kwarg):
            def _retrieve(kw):
                try:
                    r = self.dispatch_response(action, **kw)
                except Exception as e:
                    self.dispatch_onfailure(action, e, parent_response, _results, _context, **kw)
                    raise
                else:
                    if self.dispatch_ignore_status_codes(action, r) is True:
                        return False
                    if not self.dispatch_response_filter(action, r, _results, _context):
                        return False
                    return r
                    
            def _parse(r):
                is_parsable = self._is_parsable(r)
                soup = self._load_soup(r.content) if is_parsable else r.content
                extracted = self.dispatch_extractor(action, soup, r, _results, _composed, _context)
                
                results, contexts = [], []
                for result_set in self.dispatch_parser(action, r, extracted, soup, _results, _composed, _context):
                    if result_set:
                        res, ctx = result_set
                        results += pluralize(res)
                        contexts.append(ctx)

                return results, contexts, r, is_parsable
            
            if (response:= _retrieve(kwarg)) is not False:
                return  _parse(response)
        
        requests_kwargs = build_requests()
        max_workers = self.settings.get('REQUEST_MAX_WORKERS')
        if action.workers:
            with futures.ThreadPoolExecutor(min(action.workers, max_workers)) as excutor:
                processeds = excutor.map(process_response, requests_kwargs)
        else:
            processeds = map(process_response, requests_kwargs)

        rendered = False
        
        for results, contexts, response, is_parsable in filter(None, processeds):
            rendered = True
            _composed[action.name] += results
            _composed_buffer[action.name] += results
            context_one = contexts[0] if contexts else {}

            action.pipeline(results, response=response, context=_context)

            if not action.bfo or is_parsable:
                if action.follow_each:
                    for result, context in zip(results, contexts):
                        self.crawl(rest, result, response, _composed, _context=context, _composed_buffer=_composed_buffer)
                else:
                    self.crawl(rest, results, response,  _composed, _context=context_one, _composed_buffer=_composed_buffer)
            else:
                self.crawl(rest, _results, parent_response, _composed, _context=context_one, _composed_buffer=_composed_buffer)

        if not rendered and action.bfo:
            self.crawl(rest, _results, parent_response, _composed, _context=_context, _composed_buffer=_composed_buffer)
    
        if _urlorders is None:
            self.dispatch_compose(_composed)


class RequestsCrawler(CrawlMixin, RequestsBase):
    pass


class CachedRequestsCrawler(CrawlMixin, CachedRequests):
    pass



import logging, os
from logging import DEBUG, WARNING, CRITICAL, INFO, NOTSET



def get_logger(logfile=None, logger_name=None,  mode='a', loglevel=DEBUG):

    logger = logging.getLogger(logger_name or __name__)

    formatter = logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(formatter)
    logger.addHandler(streamHandler)

    if logfile:
        diranme = os.path.dirname(logfile)
        os.makedirs(diranme, exist_ok=True)
        fileHandler = logging.FileHandler(logfile,mode=mode)
        fileHandler.setFormatter(formatter)
        logger.addHandler(fileHandler)

    logger.setLevel(level=loglevel)
    return logger




from typing import *
from string import Formatter

import listorm

from ..lib import fsave, fbasename, fjoin, select_kwargs



def _get_format_kwargs(formatstr):
    formatter = Formatter()
    return [
        kw for _, kw, _, _ in formatter.parse(formatstr)
    ]



def pipe2file(basedir, subdir:Union[Callable, str]=None):
    '''
    pipe2file(RESULTS_PATH, "{titleName}({titleId})/thumbnail")
    :param subdir: lamabda response, **context: return dir, defaults to None
    '''
    if isinstance(subdir, str):
        format_args = _get_format_kwargs(subdir)
    
    def pipe(results, response, context):
        if subdir:
            if isinstance(subdir, str):
                kwargs = listorm.asselect(context, format_args)
                save_dir = subdir.format(**kwargs)
                save_path = fjoin(basedir, save_dir, fbasename(response.url))
            elif isinstance(subdir, Callable):
                save_path = select_kwargs(subdir, response, **context)
                # save_path = subdir(response, **context)
                save_path = fjoin(basedir, save_path)
            else:
                raise ValueError(subdir)
        else:
            save_path = basedir
        fsave(response.content, save_path)

    return pipe


import requests_cache
from requests import PreparedRequest

from scrapgo.lib.module import select_kwargs

from .base import RequestsBase


class CachedRequests(RequestsBase):
    cached_session_params = [
        'key_fn', 'ignored_parameters', 'match_headers', 'stale_if_error', 'filter_fn',
        'always_revalidate', 'expire_after',
    ]
    cached_request_params = [
        'refresh', 'only_if_cached', 'expire_after', 'force_refresh ',
    ]

    def load_session(self):
        return self.apply_settings(
            requests_cache.CachedSession,
            prefix='REQUEST_CACHE_',
            allowed_params=self.cached_session_params
        )
    
    def delete_cache(self, **requests_kwargs):
        if not requests_kwargs.get('request'):
            preq = PreparedRequest()
            select_kwargs(preq.prepare, **requests_kwargs)
            requests_kwargs['request'] = preq
        key = self.session.cache.create_key(**requests_kwargs)
        re = self.session.cache.delete(key)

    def fetch(self, *args, **kwargs):
        allowed_params = self.request_params + self.cached_request_params
        return super().fetch(*args, allowed_params=allowed_params, **kwargs)


from .url import *
from .regex import *
from .module import *
from .etc import *
from .parser import *
from .curl import curl2requests
from .times import *
from .plural import *
from .sh import *
from .logger import *


from .crawler import CachedRequestsCrawler, RequestsCrawler, urlpattern, urlrender, curl, file
from .core import FileRequestsMixin
from .settings import DO_NOT_CACHE, NEVER_EXPIRE, EXPIRE_IMMEDIATELY
from .pipeline import pipe2excel, pipe2jsonl, pipe2file


from typing import Union
import listorm, os, atexit, functools




def excel2context(path, context_name, asvalues:Union[str, tuple]=None, sheetname=None):
    def wrapper(renderer):
        @functools.wraps(renderer)
        def load(*args, **kwargs):
            records = listorm.read_excel(path, sheet_name=sheetname)
            if asvalues:
                records = listorm.values(records, asvalues)
            kwargs[context_name] = records
            return renderer(*args, **kwargs)
        return load
    return wrapper




def pipe2excel(path, sheet_name=None, unique_keys=None, mode='append', bulk_size=1000, image_fields=None):
    records_buffer = []
    exists_values = []
    if os.path.exists(path):
        if unique_keys:
            exists_values = listorm.values(listorm.read_excel(path), unique_keys)
            
        if mode == 'overwrite':
            with open(path, 'wb') as fp:
                fp.close()

    def _feed_excel(records):
        if unique_keys:
            records = listorm.select(records,
                where=lambda **row: listorm.asvalues(row, unique_keys) not in exists_values
            )
        print(f'pipe2excel: feed {len(records)} records to {path}...')
        listorm.write_excel(records, path, sheet_name, mode='append', image_fields=image_fields)

    def __flush_buffer():
        _feed_excel(records_buffer)

    atexit.register(__flush_buffer)

    def pipe(records, **kwargs):
        nonlocal records_buffer

        for row in records:
            records_buffer.append(row)

        buffer_size = len(records_buffer)

        if buffer_size >= bulk_size:
            _feed_excel(records_buffer)
            records_buffer = []
    return pipe
        




import jsonlines, listorm, os, functools
from typing import Union



def read_jsonl(path, **kwargs):
    if os.path.exists(path):
        with jsonlines.open(path, **kwargs) as fp:
            return list(fp)
    return []


def jsonl2context(path, context_name, asvalues:Union[str, tuple]=None):
    def wrapper(renderer):
        @functools.wraps(renderer)
        def load(*args, **kwargs):
            records = read_jsonl(path)
            if asvalues:
                records = listorm.values(records, asvalues)
            kwargs[context_name] = records
            return renderer(*args, **kwargs)
        return load
    return wrapper

    

def pipe2jsonl(file, unique_keys=None, mode='append', **jsonlines_kwargs):
    exists_values = []
    if isinstance(file, str):
        if os.path.exists(file):
            if mode == 'overwrite':
                with open(file, 'w') as fp:
                    fp.close()

            if unique_keys:
                exists_records = read_jsonl(file)
                exists_values = listorm.values(exists_records, unique_keys)
        else:
            os.makedirs(os.path.dirname(file), exist_ok=True)

    def pipe(records, **response):
        with jsonlines.open(file, mode='a', **jsonlines_kwargs) as fp:
            if unique_keys:
                records = listorm.select(
                    records,
                    where=lambda **row: listorm.asvalues(row, unique_keys) not in exists_values
                )
            for row in records:
                fp.write(row)
    return pipe


